#!perl -T

use Test::More tests => 1;

BEGIN {
    use_ok( 'Spreadsheet::WriteExcel::Extended::FitColumnWidth' );
}

diag( "Testing Spreadsheet::WriteExcel::Extended::FitColumnWidth $Spreadsheet::WriteExcel::Extended::FitColumnWidth::VERSION, Perl $], $^X" );
